import torch
import torch.nn as nn
import torchvision.transforms as transforms
import torchvision.datasets as dsets
import torchvision.models as models
import torch.nn.utils.prune as prune
from torch.autograd import Variable
import numpy as np
import os
import random
import matplotlib.pyplot as plt
import torch.nn.init as init
import copy
from matplotlib import cm, colors
from mpl_toolkits.mplot3d import Axes3D
import torch.utils.data as utils
import torch.nn.functional as F
from torch.optim.optimizer import Optimizer, required
from torch.utils.data import Dataset, DataLoader, TensorDataset
from torchvision.utils import save_image
from torch.autograd import Variable
from torch import Tensor
from torch.nn.parameter import Parameter
import general_energy_nets.predictive_coding as pc



class Assembly_Layer(nn.Module):
    def __init__(self, in_features,n_output, out_features, bias=True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.n_output = n_output

        self.PCLayer = pc.PCLayer()
        self.Linear = nn.Linear(in_features,in_features)

        linear_list = []
        for i in range(n_output):
            linear_list.append(nn.Linear(out_features,out_features))
       
        
    def forward(self, x):

        out = self.PCLayer(x)
        out = self.Linear(out)
        out = self.PCLayer(out)

        return out


class cut_layer(nn.Module):
    def __init__(self, hd):
        super().__init__()
        self.hd = hd
    def forward(self,x):
        return x[:,-self.hd:]

class uncut_layer(nn.Module):
    def __init__(self, hd,N):
        super().__init__()
        self.diff = N - hd
    
    def forward(self,x):
        torch.zeros(x.size[0], self.diff).cuda()
        return x[:,-self.hd:]


def load_dataset(dataset_name, partial, partial2, batch_size,test_batch_size,augmentation = False, shuffle = True):

    if dataset_name == 'MNIST':

        input_dim = 28*28

        train_dataset = dsets.MNIST(root='./data', 
                                    train=True, 
                                    transform=transforms.ToTensor(),
                                    download=True,
                                    )

        test_dataset = dsets.MNIST(root='./data', 
                                   train=False, 
                                   transform=transforms.ToTensor())

        


    if dataset_name == 'FashionMNIST':

        input_dim = 28*28

        train_dataset = dsets.FashionMNIST(root='./data', 
                                    train=True, 
                                    transform=transforms.ToTensor(),
                                    download=True)

        test_dataset = dsets.FashionMNIST(root='./data', 
                                   train=False, 
                                   transform=transforms.ToTensor())


    if dataset_name == 'CIFAR10':

        input_dim = 32*32*3

        train_dataset = dsets.CIFAR10(root='./data', 
                                train=True, 
                                transform=transforms.ToTensor(),
                                download=True)


        if augmentation == False:

            test_dataset = dsets.CIFAR10(root='./data', 
                                    train=False, 
                                    transform=transforms.ToTensor())

        if augmentation == True:
            transform_test = transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
            ])

            test_dataset = dsets.CIFAR10(root='./data', 
                                    train=False, 
                                    transform=transform_test)


    #input_dim = train_dataset[0][0].size()
    output_dim = 0
    for i in range(len(train_dataset)):
        if train_dataset[i][1] > output_dim:
            output_dim = train_dataset[i][1]

    output_dim = output_dim + 1

    train_list = []
    label_list = []
    test_list = []
    test_label_list = []

    if partial > 0:
        for i in range(output_dim):
            count = 0
            for j in range(len(train_dataset)):
                if  (int(train_dataset[j][1]) == i) and (count < partial):
                    train_list.append(train_dataset[j][0].view(input_dim))
                    label_list.append(torch.eye(output_dim)[train_dataset[j][1]])
                    count += 1

    else:
        for j in range(len(train_dataset)):
            train_list.append(train_dataset[j][0].view(input_dim))
            label_list.append(torch.eye(output_dim)[train_dataset[j][1]])

    if partial2 > 0:
        for i in range(output_dim):
            count2 = 0
            for k in range(len(test_dataset)):
                if  (int(test_dataset[k][1]) == i) and (count2 < partial2):
                    test_list.append(test_dataset[k][0].view(input_dim))
                    test_label_list.append(torch.eye(output_dim)[test_dataset[k][1]])
                    count2 += 1

    tensor_train = torch.stack(train_list)
    tensor_label = torch.stack(label_list)

    tensor_test = torch.stack(test_list)
    tensor_test_label = torch.stack(test_label_list)


    if augmentation == True:

        train_trans = [
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))
                ]

        my_dataset = CustomTensorDataset((tensor_train,tensor_label),transform = transforms.Compose(train_trans))
    else:
        my_dataset = TensorDataset(tensor_train,tensor_label)
        my_test_dataset = TensorDataset(tensor_test,tensor_test_label)


    train_loader = DataLoader(dataset = my_dataset,
                                batch_size=batch_size,
                                shuffle = shuffle,
                                drop_last=True,
                                )

    

    test_loader = torch.utils.data.DataLoader(dataset=my_test_dataset,
                                               batch_size=test_batch_size,
                                               shuffle=True)

        
    return train_loader, test_loader





def load_am_dataset(dataset_name,input_dim,output_dim,batch_size,line = 0,partial = 0, gaussian = 0):


    trans=transforms.Compose([
                transforms.ToTensor(),
                AddGaussianNoise(gaussian)
            ])

    if line > 0:

        trans=transforms.Compose([
            transforms.ToTensor(),
            AddPatch(1)
        ])



    if dataset_name == 'FashionMNIST':

        train_dataset = dsets.FashionMNIST(root='./data', 
                                    train=True, 
                                    transform=transforms.ToTensor(),
                                    download=True)

        noisy_dataset = dsets.FashionMNIST(root='./data', 
                                    train=True, 
                                    transform=trans,
                                    download=True)

    if dataset_name == 'MNIST':

        train_dataset = dsets.MNIST(root='./data', 
                                    train=True, 
                                    transform=transforms.ToTensor(),
                                    download=True)

        noisy_dataset = dsets.MNIST(root='./data', 
                                    train=True, 
                                    transform=trans,
                                    download=True)

    if dataset_name == 'CIFAR10':

        train_dataset = dsets.CIFAR10(root='./data', 
                            train=True, 
                            transform=transforms.ToTensor(),
                            download=False)

        noisy_dataset = dsets.CIFAR10(root='./data', 
                                train=True, 
                                transform=trans)

    train_list = []
    noisy_list = []
    label_list = []

    if partial == 0:

        for j in range(len(train_dataset)):
            label_list.append(train_dataset[j][0].view(input_dim))
            noisy_list.append(noisy_dataset[j][0].view(input_dim))
            train_list.append(1.0)

    if partial > 0:
        for i in range(output_dim):
            count = 0
            count_pre = 0
            for j in range(len(train_dataset)):
                if  (int(train_dataset[j][1]) == i) and (count < partial):
                    noisy_list.append(noisy_dataset[j][0].view(input_dim))
                    label_list.append(train_dataset[j][0].view(input_dim))
                    train_list.append(torch.tensor([1]))
                    count += 1
                if count == partial:
                    break

    tensor_train = torch.stack(train_list)
    tensor_label = torch.stack(label_list)
    tensor_noise = torch.stack(noisy_list)

    my_dataset =  TensorDataset(tensor_train,tensor_label)
    noisy_dataset = TensorDataset(tensor_train,tensor_noise)


    train_loader = DataLoader(dataset = noisy_dataset,
                                batch_size=partial*10,
                                shuffle = False,
                                drop_last=True,
                                )


    return train_loader


