import torch
import torch.nn as nn
import torch.nn.functional as F
import statistics
import torchvision.datasets as dsets
import general_energy_nets.predictive_coding as pc
import torchvision
import torch.optim as optim
import argparse
import matplotlib
import matplotlib.pyplot as plt
import torchvision.transforms as transforms
from tqdm import tqdm
from torchvision import datasets
from torch.utils.data import Dataset, DataLoader, TensorDataset
from torchvision.utils import save_image
from torch.autograd import Variable
import Utils as u
import numpy as np
from tabulate import tabulate

# Parameters of the model
input_dim = 3*32*32 #28*28
hidden_dim = 4096
output_dim = 10
fraction  = input_dim/2

# Number of iterations performed by the train and test inference, respectively.
epochs = 1
num_iterations = 1 #5000
iterations_test = 2000

# Parameters of the algorithm. 
p_lr = 1e-5
x_lr = 0.5
xlr_test = 0.1
dataset_name = 'CIFAR10'


device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

partial = 5
batch_size = 50

if 'MNIST' in dataset_name:
    fraction = 392
    input_shape = (-1,1,28,28)
if 'CIFAR' in dataset_name:
    fraction = 512
    input_shape = (-1,3,32,32)


#################################################################################

def reconstruction(model, dataloader):
    def callback_reconstruction(t, pc_trainer, img, labels):
        model.PClabel.get_x().data.copy_(labels)
        inputs = model.PCimg.get_x().data
        if 'CIFAR' in dataset_name:
            model.PCimg.get_x().data[:,:fraction].copy_(lab_b)
            model.PCimg.get_x().data[:,1024:1024+fraction].copy_(lab_r)
            model.PCimg.get_x().data[:,2048:2048+fraction].copy_(lab_g)
        if 'MNIST' in dataset_name:
            model.PCimg.get_x().data[:,:fraction].copy_(lab)
    
    loss = 0.0

    for i, (images, labels) in enumerate(dataloader):
        model.train()
        num_rows = 10
        callback_after_t_kwargs={'pc_trainer':pc_trainer,'img':images,'labels':labels}
        if 'CIFAR' in dataset_name:
            lab_b = images[:,:fraction]
            lab_r = images[:,1024:1024+fraction]
            lab_g = images[:,2048:2048+fraction]
        if 'MNIST' in dataset_name:
            lab = images[:,:fraction]
            #print(lab.shape)


        if 'CIFAR' in dataset_name:
            zeros = torch.zeros((batch_size,3072))
            zeros[:,:fraction] = lab_b
            zeros[:,1024:1024+fraction] = lab_r
            zeros[:,2048:2048+fraction] = lab_g
        if 'MNIST' in dataset_name:
            zeros = torch.zeros((batch_size,784))
            zeros[:,:fraction] = lab

        images = Variable(images.view(-1,1).cuda())
        fake_inputs = torch.rand(batch_size,3*32*32).cuda()
        results = prediction_trainer.train_on_batch(fake_inputs.float(),is_log_progress=False,callback_after_t=callback_reconstruction,callback_after_t_kwargs=callback_after_t_kwargs)
        loss += results['overall'][-1]
        reconstruction = model.PCimg.get_x().data.cpu().view(batch_size,-1)[:,:input_dim]
        noise = images.view(batch_size,-1).cpu() - reconstruction
        if partial > 0: 
            image = torch.cat((zeros.view(input_shape)[0::partial], 
                                reconstruction.view(input_shape)[0::partial],
                                images.view(input_shape)[0::partial].cpu(),
                                noise.view(input_shape)[0::partial]))
            save_image(image.cpu(), f"./bi_cpc/reconstruction0821_cifar/-{dataset_name} - L= {3},hidden_dim={hidden_dim},partial={partial},frac={fraction},T={num_iterations},T_recon={iterations_test},xlr={x_lr},plr={p_lr}.png", nrow=10)
        break

###########################################################################################

class Bidirectional_Model(nn.Module):
    def __init__(self, input_dim: int, hidden_dim: int, output_dim:int) -> None:
        super(Bidirectional_Model, self).__init__()

        self.PCimg = pc.PCLayer()
        self.PClabel = pc.PCLayer()

        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, hidden_dim)

        self.fc4 = nn.Linear(hidden_dim, output_dim)
        self.fc4inv = nn.Linear(output_dim, hidden_dim)

        self.fc3inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc2inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc1inv = nn.Linear(hidden_dim, input_dim)

        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.relu3 = nn.ReLU()

        self.relu4 = nn.ReLU()

        self.PClayer1 = pc.PCLayer()
        self.PClayer2 = pc.PCLayer()
        self.PClayer3 = pc.PCLayer()

    def forward(self, x):

        out = self.PCimg(x)

        out = self.fc1(out)
        out = self.PClayer1(out)
        out = self.relu1(out)

        out = self.fc2(out)
        out = self.PClayer2(out)
        out = self.relu2(out)

        out = self.fc3(out)
        out = self.PClayer3(out)
        out = self.relu3(out)

        label = self.fc4(out)
        out = self.PClabel(label)
        out = self.fc4inv(out)
    
        out = self.relu3(out)
        out = self.PClayer3(out)
        out = self.fc3inv(out)

        out = self.relu2(out)
        out = self.PClayer2(out)
        out = self.fc2inv(out)

        out = self.relu1(out) 
        out = self.PClayer1(out)  
        out = self.fc1inv(out)
        out = self.PCimg(out)

        return label

########################################################################
def callback_after_t(t, pc_trainer,img,labels):
    model.PClabel.get_x().data.copy_(labels.cuda())
    model.PCimg.get_x().data.copy_(img.cuda())


def loss_fn(output):
    return (output - labels.cuda()).pow(2).sum() * 0.5

T_list = [1]#1000,100000,1000, 5000, 10000, 10000 
p_lr_list = [1e-5]
hidden_dim_list = [4096]#2048,3064,4096, 5120, 
partial_list = [2,5,10]#2,
grid = [(q, x, y, z)  for q in partial_list for x in T_list for y in p_lr_list for z in hidden_dim_list]

for partial, num_iterations, p_lr, hidden_dim in grid:
    batch_size = partial*10
    test_batch_size = batch_size
    train_loader, test_loader = u.load_dataset(dataset_name, partial, batch_size,test_batch_size, augmentation = False, shuffle = True)
    print(train_loader.dataset.tensors[0].data.shape)

    model = Bidirectional_Model(input_dim,hidden_dim,output_dim)
    model.cuda()
    iterations_test = 250
    print("parameters are: ",partial, num_iterations, p_lr, hidden_dim)
    pc_trainer = pc.PCTrainer(model,optimizer_x_kwargs={'lr': x_lr}, optimizer_p_kwargs={'lr': p_lr}, T = num_iterations, early_stop_condition="True")
    prediction_trainer = pc.PCTrainer(model, optimizer_x_kwargs={'lr': xlr_test}, optimizer_p_kwargs={'lr': 0}, T = iterations_test,early_stop_condition="True")
    for epoch in range(epochs):
        model.train()
        for i, (data, labels) in enumerate(train_loader):

            callback_after_t_kwargs={'pc_trainer':pc_trainer,'img':data,'labels':labels}
            pc_trainer.train_on_batch(data.cuda(),loss_fn, is_log_progress=True,callback_after_t=callback_after_t,callback_after_t_kwargs=callback_after_t_kwargs)
        reconstruction(model,train_loader)
        del pc_trainer
        del prediction_trainer